@@ -24,8 +24,6 @@
 #include "gdm.h"
 #include "filecheck.h"
 
-static const gchar RCSid[]="$Id: filecheck.c,v 1.5 2001/03/05 00:29:54 jirka Exp $";
-
 /**
  * gdm_file_check:
  * @caller: String to be prepended to syslog error messages.
@@ -81,7 +79,7 @@ gdm_file_check (const gchar *caller, uid_t user, const gchar *dir,
 	    return TRUE;
 	}
 	else {
-	    syslog (LOG_WARNING, _("%s: %s does not exist and must."), caller, fullpath);
+	    syslog (LOG_WARNING, _("%s: %s does not exist but must exist."), caller, fullpath);
 	    g_free (fullpath);
 	    return FALSE;
 	}
