@@ -29,15 +29,14 @@
 #include <arpa/inet.h>
 #include <netdb.h>
 #include <errno.h>
+#include <grp.h>
+#include <sys/types.h>
 
 #include <vicious.h>
 
 #include "gdm.h"
 #include "misc.h"
 
-static const gchar RCSid[]="$Id: misc.c,v 1.28 2001/10/28 19:13:42 jirka Exp $";
-
-
 /* Configuration option variables */
 extern gchar *GdmPidFile;
 extern gboolean GdmDebug;
@@ -283,7 +282,8 @@ gdm_text_message_dialog (const char *msg)
 		argv[6] = NULL;
 
 		/* make sure gdialog wouldn't get confused */
-		if (gdm_exec_wait (argv, TRUE /* no display */) < 0) {
+		if (gdm_exec_wait (argv, TRUE /* no display */,
+				   TRUE /* de_setuid */) < 0) {
 			g_free (dialog);
 			return FALSE;
 		}
@@ -301,7 +301,8 @@ gdm_text_message_dialog (const char *msg)
 			 msg);
 		argv[4] = NULL;
 
-		if (gdm_exec_wait (argv, TRUE /* no display */) < 0) {
+		if (gdm_exec_wait (argv, TRUE /* no display */,
+				   TRUE /* de_setuid */) < 0) {
 			g_free (argv[3]);
 			return FALSE;
 		}
@@ -340,7 +341,8 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 
 		/* will unset DISPLAY and XAUTHORITY if they exist
 		 * so that gdialog (if used) doesn't get confused */
-		retint = gdm_exec_wait (argv, TRUE /* no display */);
+		retint = gdm_exec_wait (argv, TRUE /* no display */,
+					TRUE /* de_setuid */);
 		if (retint < 0) {
 			g_free (dialog);
 			return FALSE;
@@ -378,7 +380,8 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 			 tempname);
 		argv[4] = NULL;
 
-		if (gdm_exec_wait (argv, TRUE /* no display */) < 0) {
+		if (gdm_exec_wait (argv, TRUE /* no display */,
+				   TRUE /* de_setuid */) < 0) {
 			g_free (argv[3]);
 			return FALSE;
 		}
@@ -403,7 +406,8 @@ gdm_text_yesno_dialog (const char *msg, gboolean *ret)
 }
 
 int
-gdm_exec_wait (char * const *argv, gboolean no_display)
+gdm_exec_wait (char * const *argv, gboolean no_display,
+	       gboolean de_setuid)
 {
 	int status;
 	pid_t pid;
@@ -426,6 +430,11 @@ gdm_exec_wait (char * const *argv, gboolean no_display)
 		open ("/dev/null", O_RDWR); /* open stdout - fd 1 */
 		open ("/dev/null", O_RDWR); /* open stderr - fd 2 */
 
+		if (de_setuid) {
+			seteuid (getuid ());
+			setegid (getgid ());
+		}
+
 		if (no_display) {
 			ve_unsetenv ("DISPLAY");
 			ve_unsetenv ("XAUTHORITY");
@@ -713,5 +722,22 @@ gdm_is_loopback_addr (struct in_addr *ia)
 	}
 }
 
+gboolean
+gdm_setup_gids (const char *login, gid_t gid)
+{
+	if (setgid (gid) < 0)  {
+		gdm_error (_("Could not setgid %d. Aborting."), (int)gid);
+		return FALSE;
+	}
+
+	if (initgroups (login, gid) < 0) {
+		gdm_error (_("initgroups() failed for %s. Aborting."), login);
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+
 
 /* EOF */
