@@ -362,7 +362,7 @@ gdm_verify_user (GdmDisplay *d,
 	    gdm_error (_("Authentication token change failed for user %s"), login);
 	    gdm_slave_greeter_ctl_no_ret (GDM_ERRBOX, 
 		    _("\nThe change of the authentication token failed. "
-		      "Please try again later or cantact the system administrator."));
+		      "Please try again later or contact the system administrator."));
 	    error_msg_given = TRUE;
 	    goto pamerr;
 	}
@@ -429,7 +429,8 @@ gdm_verify_user (GdmDisplay *d,
     if ( ! error_msg_given &&
 	gdm_slave_should_complain ()) {
 	    /* I'm not sure yet if I should display this message for any other issues - heeten */
-	    if (pamerr == PAM_AUTH_ERR) {
+	    if (pamerr == PAM_AUTH_ERR ||
+		pamerr == PAM_USER_UNKNOWN) {
 		    /* FIXME: Hmm, how are we sure that the login is username
 		     * and password.  That is the most common case but not
 		     * neccessairly true, this message needs to be changed
@@ -532,7 +533,7 @@ gdm_verify_setup_user (GdmDisplay *d, const gchar *login, const gchar *display)
 	gdm_error (_("User %s not permitted to gain access at this time"), login);
 	gdm_error_box (cur_gdm_disp,
 		       GNOME_MESSAGE_BOX_ERROR,
-		       _("\nThe system administrator has your disabled access to the system temporary."));
+		       _("\nThe system administrator has disabled your access to the system temporary."));
 	goto setup_pamerr;
     default :
 	if (gdm_slave_should_complain ())
