@@ -59,6 +59,7 @@
 
 #ifdef HAVE_LIBXDMCP
 #include <stdio.h>
+#include <time.h>
 #include <stdlib.h>
 #include <syslog.h>
 #include <sys/stat.h>
@@ -84,6 +85,8 @@
   #include <tcpd.h>
 #endif
 
+#include <vicious.h>
+
 #include "gdm.h"
 #include "display.h"
 #include "auth.h"
@@ -93,7 +96,6 @@
 #include "choose.h"
 #include "xdmcp.h"
 
-static const gchar RCSid[]="$Id: xdmcp.c,v 1.28 2001/07/03 09:13:33 jirka Exp $";
 gint pending = 0;
 
 #ifdef HAVE_LIBXDMCP
@@ -102,10 +104,8 @@ gint pending = 0;
 gint allow_severity = LOG_INFO;
 gint deny_severity = LOG_WARNING;
 
-static gint xdmcpfd = -1;
+static gint gdm_xdmcpfd = -1;
 static guint xdmcp_source = 0;
-static gint choosefd = -1;
-static guint choose_source = 0;
 static gint globsessid;
 static gchar *sysid;
 static ARRAY8 servhost;
@@ -126,6 +126,7 @@ extern gint GdmMaxIndirectWait;	/* Max wait between INDIRECT_QUERY and MANAGE */
 extern gint GdmDispPerHost;	/* Max number of displays per remote host */
 extern gchar *GdmTimedLogin;
 extern gboolean GdmAllowRemoteAutoLogin;
+extern gchar *GdmWilling;	/* The willing script */
 
 extern gboolean GdmXdmcp;	/* xdmcp enabled */
 
@@ -138,24 +139,50 @@ static void gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa,
 				    gint type);
 static void gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id,
 					  struct sockaddr_in *clnt_sa,
-					  struct in_addr *local_addr,
+					  struct in_addr *display_addr,
 					  ARRAYofARRAY8Ptr authlist);
 static void gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len);
 static void gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len);
 static void gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len);
+static void gdm_xdmcp_handle_managed_forward (struct sockaddr_in *clnt_sa, gint len);
+static void gdm_xdmcp_handle_got_managed_forward (struct sockaddr_in *clnt_sa, gint len);
 static void gdm_xdmcp_handle_keepalive (struct sockaddr_in *clnt_sa, gint len);
+static void gdm_xdmcp_whack_queued_managed_forwards (struct sockaddr_in *clnt_sa,
+						     struct in_addr *origin);
 static void gdm_xdmcp_send_willing (struct sockaddr_in *clnt_sa);
 static void gdm_xdmcp_send_unwilling (struct sockaddr_in *clnt_sa, gint type);
 static void gdm_xdmcp_send_accept (const char *hostname, struct sockaddr_in *clnt_sa, gint displaynum);
-static void gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa);
+static void gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa, const char *reason);
 static void gdm_xdmcp_send_refuse (struct sockaddr_in *clnt_sa, CARD32 sessid);
 static void gdm_xdmcp_send_failed (struct sockaddr_in *clnt_sa, CARD32 sessid);
 static void gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD32 sessid);
+static void gdm_xdmcp_send_managed_forward (struct sockaddr_in *clnt_sa,
+					    struct sockaddr_in *origin);
+static void gdm_xdmcp_send_got_managed_forward (struct sockaddr_in *clnt_sa,
+						struct in_addr *origin);
 static gboolean gdm_xdmcp_host_allow (struct sockaddr_in *cnlt_sa);
-static GdmDisplay *gdm_xdmcp_display_alloc (const char *hostname, gint);
+static GdmDisplay *gdm_xdmcp_display_alloc (struct in_addr *addr, const char *hostname, gint);
 static GdmDisplay *gdm_xdmcp_display_lookup (CARD32 sessid);
 static void gdm_xdmcp_display_dispose_check (const gchar *name);
 static void gdm_xdmcp_displays_check (void);
+static int gdm_xdmcp_displays_from_host (struct in_addr *addr);
+
+static GdmForwardQuery * gdm_forward_query_alloc (struct sockaddr_in *mgr_sa,
+						  struct sockaddr_in *dsp_sa);
+static GdmForwardQuery * gdm_forward_query_lookup (struct sockaddr_in *clnt_sa);
+static void gdm_forward_query_dispose (GdmForwardQuery *q);
+
+static GSList *forward_queries = NULL;
+
+typedef struct {
+	int times;
+	guint handler;
+	struct sockaddr_in manager; 
+	struct sockaddr_in origin;
+} ManagedForward;
+#define MANAGED_FORWARD_INTERVAL 1500 /* 1.5 seconds */
+
+static GList *managed_forwards = NULL;
 
 
 /* 
@@ -180,143 +207,23 @@ static XdmAuthRec serv_authlist = {
     { (CARD16) 0, (CARD8 *) 0 }
 };
 
-static const GList *
-peek_local_address_list (void)
+static int
+gdm_xdmcp_displays_from_host (struct in_addr *addr)
 {
-	static GList *the_list = NULL;
-	static time_t last_time = 0;
-	struct in_addr *addr;
-#ifdef SIOCGIFCONF
-	struct sockaddr_in *sin;
-	struct ifconf ifc;
-	struct ifreq *ifr;
-	char *buf;
-	int num;
-#else /* SIOCGIFCONF */
-	char hostbuf[BUFSIZ];
-	struct hostent *he;
-	int i;
-#endif
+	GSList *li;
+	int count = 0;
 
-	/* don't check more then every 5 seconds */
-	if (last_time + 5 > time (NULL))
-		return the_list;
-
-	g_list_foreach (the_list, (GFunc)g_free, NULL);
-	g_list_free (the_list);
-	the_list = NULL;
-
-	last_time = time (NULL);
-
-#ifdef SIOCGIFCONF
-#ifdef SIOCGIFNUM
-	if (ioctl (xdmcpfd, SIOCGIFNUM, &num) < 0) {
-		num = 64;
-	}
-#else
-	num = 64;
-#endif
-
-	ifc.ifc_len = sizeof(struct ifreq) * num;
-	ifc.ifc_buf = buf = g_malloc (ifc.ifc_len);
-	if (ioctl (xdmcpfd, SIOCGIFCONF, &ifc) < 0) {
-		gdm_error (_("%s: Cannot get local addresses!"),
-			   "peek_local_address_list");
-		return NULL;
+	for (li = displays; li != NULL; li = li->next) {
+		GdmDisplay *disp = li->data;
+		if (disp->type == TYPE_XDMCP &&
+		    memcmp (&disp->addr, addr, sizeof (struct in_addr)) == 0)
+			count ++;
 	}
 
-	ifr = ifc.ifc_req;
-	num = ifc.ifc_len / sizeof(struct ifreq);
-	for (; num-- > 0; ifr++) {
-		if (ioctl (xdmcpfd, SIOCGIFFLAGS, ifr) < 0 ||
-		    ! (ifr->ifr_flags & IFF_UP))
-			continue;
-#ifdef IFF_UNNUMBERED
-		if (ifr->ifr_flags & IFF_UNNUMBERED)
-			continue;
-#endif
-		if (ioctl (xdmcpfd, SIOCGIFADDR, ifr) < 0)
-			continue;
-
-		sin = (struct sockaddr_in *)&ifr->ifr_addr;
-
-		if (sin->sin_family != AF_INET ||
-		    sin->sin_addr.s_addr == INADDR_ANY ||
-		    sin->sin_addr.s_addr == INADDR_LOOPBACK)
-			continue;
-		addr = g_new0 (struct in_addr, 1);
-		memcpy (addr, &(sin->sin_addr.s_addr),
-			sizeof (struct in_addr));
-		the_list = g_list_append (the_list, addr);
-	}
-
-	g_free (buf);
-#else /* SIOCGIFCONF */
-	/* host based fallback, will likely only get 127.0.0.1 i think */
-
-	if (gethostname (hostbuf, BUFSIZ-1) != 0) {
-		gdm_error (_("%s: Could not get server hostname: %s!"),
-			   "peek_local_address_list",
-			   strerror (errno));
-		return NULL;
-	}
-	he = gethostbyname (hostbuf);
-	if (he == NULL) {
-		gdm_error (_("%s: Could not get address from hostname!"),
-			   "peek_local_address_list");
-		return NULL;
-	}
-	for (i = 0; he->h_addr_list[i] != NULL; i++) {
-		struct in_addr *laddr = (struct in_addr *)he->h_addr_list[i];
-
-		addr = g_new0 (struct in_addr, 1);
-		memcpy (addr, laddr, sizeof (struct in_addr));
-		the_list = g_list_append (the_list, addr);
-	}
-#endif
-
-	return the_list;
+	return count;
 }
 
 
-static gboolean
-is_loopback_addr (struct in_addr *ia)
-{
-	const char lo[] = {127,0,0,1};
-
-	if (ia->s_addr == INADDR_LOOPBACK ||
-	    memcmp (&ia->s_addr, lo, 4) == 0) {
-		return TRUE;
-	} else {
-		return FALSE;
-	}
-}
-
-static gboolean
-is_local_addr (struct in_addr *ia)
-{
-	const char lo[] = {127,0,0,1};
-
-	if (ia->s_addr == INADDR_LOOPBACK ||
-	    memcmp (&ia->s_addr, lo, 4) == 0) {
-		return TRUE;
-	} else {
-		const GList *list = peek_local_address_list ();
-
-		while (list != NULL) {
-			struct in_addr *addr = list->data;
-
-			if (memcmp (&ia->s_addr, &addr->s_addr, 4) == 0) {
-				return TRUE;
-			}
-
-			list = list->next;
-		}
-
-		return FALSE;
-	}
-}
-
 gboolean
 gdm_xdmcp_init (void)
 {
@@ -342,9 +249,9 @@ gdm_xdmcp_init (void)
     gdm_debug ("Start up on host %s, port %d", hostbuf, GdmPort);
     
     /* Open socket for communications */
-    xdmcpfd = socket (AF_INET, SOCK_DGRAM, 0); /* UDP */
+    gdm_xdmcpfd = socket (AF_INET, SOCK_DGRAM, 0); /* UDP */
     
-    if (xdmcpfd < 0) {
+    if (gdm_xdmcpfd < 0) {
 	gdm_error (_("gdm_xdmcp_init: Could not create socket!"));
 	GdmXdmcp = FALSE;
 	return FALSE;
@@ -354,45 +261,14 @@ gdm_xdmcp_init (void)
     serv_sa.sin_port = htons (GdmPort); /* UDP 177 */
     serv_sa.sin_addr.s_addr = htonl (INADDR_ANY);
     
-    if (bind (xdmcpfd, (struct sockaddr_in *) &serv_sa, sizeof (serv_sa)) == -1) {
+    if (bind (gdm_xdmcpfd, (struct sockaddr*) &serv_sa, sizeof (serv_sa)) == -1) {
 	gdm_error (_("gdm_xdmcp_init: Could not bind to XDMCP socket!"));
 	gdm_xdmcp_close ();
 	GdmXdmcp = FALSE;
 	return FALSE;
     }
 
-    /* Setup FIFO if choosing is enabled */
-    if (GdmIndirect) {
-	gchar *fifopath;
-
-	fifopath = g_strconcat (GdmServAuthDir, "/.gdmchooser", NULL);
-
-	unlink (fifopath);
-
-	if (mkfifo (fifopath, 0660) < 0) {
-	    gdm_error (_("gdm_xdmcp_init: Could not make FIFO for chooser"));
-	    gdm_xdmcp_close ();
-	    GdmXdmcp = FALSE;
-	    g_free (fifopath);
-	    return FALSE;
-	}
-
-	choosefd = open (fifopath, O_RDWR); /* Open with write to avoid EOF */
-
-	if (choosefd < 0) {
-	    gdm_error (_("gdm_xdmcp_init: Could not open FIFO for chooser"));
-	    gdm_xdmcp_close ();
-	    GdmXdmcp = FALSE;
-	    g_free (fifopath);
-	    return FALSE;
-	}
-
-	chmod (fifopath, 0660);
-
-	g_free (fifopath);
-    }
-
-    return (TRUE);
+    return TRUE;
 }
 
 
@@ -401,23 +277,12 @@ gdm_xdmcp_run (void)
 {
 	GIOChannel *xdmcpchan;
 
-	xdmcpchan = g_io_channel_unix_new (xdmcpfd);
+	xdmcpchan = g_io_channel_unix_new (gdm_xdmcpfd);
 	xdmcp_source = g_io_add_watch_full
 		(xdmcpchan, G_PRIORITY_DEFAULT,
 		 G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
 		 gdm_xdmcp_decode_packet, NULL, NULL);
 	g_io_channel_unref (xdmcpchan);
-
-	if (GdmIndirect) {
-		GIOChannel *choosechan;
-
-		choosechan = g_io_channel_unix_new (choosefd);
-		choose_source = g_io_add_watch_full
-			(choosechan, G_PRIORITY_DEFAULT,
-			 G_IO_IN|G_IO_PRI|G_IO_ERR|G_IO_HUP|G_IO_NVAL,
-			 gdm_choose_socket_handler, NULL, NULL);
-		g_io_channel_unref (choosechan);
-	}
 }
 
 
@@ -429,19 +294,9 @@ gdm_xdmcp_close (void)
 		xdmcp_source = 0;
 	}
 
-	if (xdmcpfd > 0) {
-		close (xdmcpfd);
-		xdmcpfd = -1;
-	}
-
-	if (choose_source > 0) {
-		g_source_remove (choose_source);
-		choose_source = 0;
-	}
-
-	if (choosefd > 0) {
-		close (choosefd);
-		choosefd = -1;
+	if (gdm_xdmcpfd > 0) {
+		close (gdm_xdmcpfd);
+		gdm_xdmcpfd = -1;
 	}
 }
 
@@ -458,8 +313,11 @@ gdm_xdmcp_decode_packet (GIOChannel *source, GIOCondition cond, gpointer data)
 	"WILLING", "UNWILLING", "REQUEST", "ACCEPT", "DECLINE", "MANAGE", "REFUSE",
 	"FAILED", "KEEPALIVE", "ALIVE"
     };
+    static const char * const gdm_opcode_names[] = {
+	"MANAGED_FORWARD", "GOT_MANAGED_FORWARD"
+    };
     
-    if (!XdmcpFill (xdmcpfd, &buf, &clnt_sa, &sa_len)) {
+    if (!XdmcpFill (gdm_xdmcpfd, &buf, (XdmcpNetaddr)&clnt_sa, &sa_len)) {
 	gdm_error (_("gdm_xdmcp_decode: Could not create XDMCP buffer!"));
 	return TRUE;
     }
@@ -469,7 +327,8 @@ gdm_xdmcp_decode_packet (GIOChannel *source, GIOCondition cond, gpointer data)
 	return TRUE;
     }
     
-    if (header.version != XDM_PROTOCOL_VERSION) {
+    if (header.version != XDM_PROTOCOL_VERSION &&
+	header.version != GDM_XDMCP_PROTOCOL_VERSION) {
 	gdm_error (_("gdm_xdmcp_decode: Incorrect XDMCP version!"));
 	return TRUE;
     }
@@ -477,6 +336,11 @@ gdm_xdmcp_decode_packet (GIOChannel *source, GIOCondition cond, gpointer data)
     if (header.opcode <= ALIVE)
 	    gdm_debug ("gdm_xdmcp_decode: Received opcode %s from client %s", 
 		       opcode_names[header.opcode], inet_ntoa (clnt_sa.sin_addr));
+    if (header.opcode >= GDM_XDMCP_FIRST_OPCODE &&
+        header.opcode < GDM_XDMCP_LAST_OPCODE)
+	    gdm_debug ("gdm_xdmcp_decode: Received opcode %s from client %s", 
+		       gdm_opcode_names[header.opcode - GDM_XDMCP_FIRST_OPCODE],
+		       inet_ntoa (clnt_sa.sin_addr));
 
     switch (header.opcode) {
 	
@@ -508,6 +372,14 @@ gdm_xdmcp_decode_packet (GIOChannel *source, GIOCondition cond, gpointer data)
     case KEEPALIVE:
 	gdm_xdmcp_handle_keepalive (&clnt_sa, header.length);
 	break;
+
+    case GDM_XDMCP_MANAGED_FORWARD:
+	gdm_xdmcp_handle_managed_forward (&clnt_sa, header.length);
+	break;
+
+    case GDM_XDMCP_GOT_MANAGED_FORWARD:
+	gdm_xdmcp_handle_got_managed_forward (&clnt_sa, header.length);
+	break;
 	
     default:
 	gdm_error (_("gdm_xdmcp_decode_packet: Unknown opcode from host %s"),
@@ -535,7 +407,8 @@ gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type)
     
     /* Crude checksumming */
     for (i = 0 ; i < clnt_authlist.length ; i++) {
-	gdm_debug ("gdm_xdmcp_handle_query: authlist: %s", clnt_authlist.data);
+	gdm_debug ("gdm_xdmcp_handle_query: authlist: %s",
+		   (char *)clnt_authlist.data);
 	explen += 2+clnt_authlist.data[i].length;
     }
     
@@ -559,21 +432,21 @@ gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type)
 		if (id != NULL &&
 		    id->chosen_host != NULL) {
 			/* if user chose us, then just send willing */
-			if (is_local_addr (id->chosen_host)) {
+			if (gdm_is_local_addr (id->chosen_host)) {
 				/* get rid of indirect, so that we don't get
 				 * the chooser */
 				gdm_choose_indirect_dispose (id);
 				gdm_xdmcp_send_willing (clnt_sa);
-			} else if (is_loopback_addr (&(clnt_sa->sin_addr))) {
+			} else if (gdm_is_loopback_addr (&(clnt_sa->sin_addr))) {
 				/* woohoo! fun, I have no clue how to get
 				 * the correct ip, SO I just send forward
 				 * queries with all the different IPs */
-				const GList *list = peek_local_address_list ();
+				const GList *list = gdm_peek_local_address_list ();
 
 				while (list != NULL) {
 					struct in_addr *addr = list->data;
 					
-					if ( ! is_loopback_addr (addr)) {
+					if ( ! gdm_is_loopback_addr (addr)) {
 						/* forward query to
 						 * chosen host */
 						gdm_xdmcp_send_forward_query
@@ -614,7 +487,7 @@ gdm_xdmcp_handle_query (struct sockaddr_in *clnt_sa, gint len, gint type)
 static void
 gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id,
 			      struct sockaddr_in *clnt_sa,
-			      struct in_addr *local_addr,
+			      struct in_addr *display_addr,
 			      ARRAYofARRAY8Ptr authlist)
 {
 	struct sockaddr_in sock = {0};
@@ -629,7 +502,7 @@ gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id,
 	gdm_debug ("gdm_xdmcp_send_forward_query: Sending forward query to %s",
 		   inet_ntoa (*id->chosen_host));
 	gdm_debug ("gdm_xdmcp_send_forward_query: Query contains %s:%d", 
-		   inet_ntoa (*local_addr),
+		   inet_ntoa (*display_addr),
 		   (int) ntohs (clnt_sa->sin_port));
 
 	authlen = 1;
@@ -644,7 +517,7 @@ gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id,
 
 	address.length = sizeof (struct in_addr);
 	address.data = (void *)g_new (struct in_addr, 1);
-	memcpy (address.data, local_addr, sizeof (struct in_addr));
+	memcpy (address.data, display_addr, sizeof (struct in_addr));
 
 	header.opcode = (CARD16) FORWARD_QUERY;
 	header.length = authlen;
@@ -660,13 +533,117 @@ gdm_xdmcp_send_forward_query (GdmIndirectDisplay *id,
 	sock.sin_family = AF_INET;
 	sock.sin_port = htons (XDM_UDP_PORT);
 	sock.sin_addr.s_addr = id->chosen_host->s_addr;
-	XdmcpFlush (xdmcpfd, &buf, (XdmcpNetaddr) &sock,
+	XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr) &sock,
 		    (int)sizeof (struct sockaddr_in));
 
 	g_free (port.data);
 	g_free (address.data);
 }
 
+static gboolean
+remove_oldest_forward (void)
+{
+	GSList *li;
+	GdmForwardQuery *oldest = NULL;
+
+	for (li = forward_queries; li != NULL; li = li->next) {
+		GdmForwardQuery *query = li->data;
+
+		if (oldest == NULL ||
+		    query->acctime < oldest->acctime) {
+			oldest = query;
+		}
+	}
+
+	if (oldest != NULL) {
+		gdm_forward_query_dispose (oldest);
+		return TRUE;
+	} else {
+		return FALSE;
+	}
+}
+
+static GdmForwardQuery *
+gdm_forward_query_alloc (struct sockaddr_in *mgr_sa,
+			 struct sockaddr_in *dsp_sa)
+{
+	GdmForwardQuery *q;
+	int count;
+
+	count = g_slist_length (forward_queries);
+
+	while (count > GDM_MAX_FORWARD_QUERIES &&
+	       remove_oldest_forward ())
+		count --;
+
+	q = g_new0 (GdmForwardQuery, 1);
+	q->dsp_sa = g_new0 (struct sockaddr_in, 1);
+	memcpy (q->dsp_sa, dsp_sa, sizeof (struct sockaddr_in));
+	q->from_sa = g_new0 (struct sockaddr_in, 1);
+	memcpy (q->from_sa, mgr_sa, sizeof (struct sockaddr_in));
+
+	forward_queries = g_slist_prepend (forward_queries, q);
+
+	return q;
+}
+
+static GdmForwardQuery *
+gdm_forward_query_lookup (struct sockaddr_in *clnt_sa)
+{
+	GSList *li, *qlist;
+	GdmForwardQuery *q;
+	time_t curtime = time (NULL);
+
+	qlist = g_slist_copy (forward_queries);
+
+	for (li = qlist; li != NULL; li = li->next) {
+		q = (GdmForwardQuery *) li->data;
+		if (q == NULL)
+			continue;
+
+		if (q->dsp_sa->sin_addr.s_addr == clnt_sa->sin_addr.s_addr) {
+			g_slist_free (qlist);
+			return q;
+		}
+
+		if (q->acctime > 0 &&
+		    curtime > q->acctime + GDM_FORWARD_QUERY_TIMEOUT)	{
+			gdm_debug ("gdm_forward_query_lookup: Disposing stale forward query from %s",
+				   inet_ntoa (clnt_sa->sin_addr));
+			gdm_forward_query_dispose (q);
+		}
+
+	}
+	g_slist_free (qlist);
+
+	gdm_debug ("gdm_forward_query_lookup: Host %s not found", 
+		   inet_ntoa (clnt_sa->sin_addr));
+
+	return NULL;
+}
+
+
+static void
+gdm_forward_query_dispose (GdmForwardQuery *q)
+{
+	if (q == NULL)
+		return;
+
+	forward_queries = g_slist_remove (forward_queries, q);
+
+	q->acctime = 0;
+
+	gdm_debug ("gdm_forward_query_dispose: Disposing %s", 
+		   inet_ntoa (q->dsp_sa->sin_addr));
+
+	g_free (q->dsp_sa);
+	q->dsp_sa = NULL;
+	g_free (q->from_sa);
+	q->from_sa = NULL;
+
+	g_free (q);
+}
+
 
 static void 
 gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
@@ -704,7 +681,8 @@ gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
     explen += 2+clnt_port.length;
     
     for (i = 0 ; i < clnt_authlist.length ; i++) {
-	gdm_debug ("gdm_xdmcp_handle_forward_query: authlist: %s", clnt_authlist.data);
+	gdm_debug ("gdm_xdmcp_handle_forward_query: authlist: %s",
+		   (char *)clnt_authlist.data);
 	explen += 2+clnt_authlist.data[i].length;
     }
     
@@ -719,6 +697,10 @@ gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
 	    goto out;
     }
 
+    g_assert (4 == sizeof (struct in_addr));
+    gdm_xdmcp_whack_queued_managed_forwards
+	    (clnt_sa, (struct in_addr *)clnt_addr.data);
+
     disp_sa.sin_family = AF_INET;
 
     /* Find client port number */
@@ -731,8 +713,16 @@ gdm_xdmcp_handle_forward_query (struct sockaddr_in *clnt_sa, gint len)
 	       inet_ntoa (disp_sa.sin_addr), ntohs (disp_sa.sin_port));
     
     /* Check with tcp_wrappers if display is allowed to access */
-    if (gdm_xdmcp_host_allow (&disp_sa)) 
+    if (gdm_xdmcp_host_allow (&disp_sa)) {
+	    GdmForwardQuery *q;
+
+	    q = gdm_forward_query_lookup (&disp_sa);
+	    if (q != NULL)
+		    gdm_forward_query_dispose (q);
+	    gdm_forward_query_alloc (clnt_sa, &disp_sa);
+
 	    gdm_xdmcp_send_willing (&disp_sa);
+    }
 
   out:
     /* Cleanup */
@@ -747,11 +737,46 @@ gdm_xdmcp_send_willing (struct sockaddr_in *clnt_sa)
 {
     ARRAY8 status;
     XdmcpHeader header;
+    static char *last_status = NULL;
+    static time_t last_willing = 0;
+    char *bin;
+    FILE *fd;
     
     gdm_debug ("gdm_xdmcp_send_willing: Sending WILLING to %s", inet_ntoa (clnt_sa->sin_addr));
-    
-    status.data = sysid;
-    status.length = strlen (sysid);
+
+    if (last_willing == 0 ||
+	time (NULL) - 3 > last_willing) {
+	    char statusBuf[256] = "";
+	    bin = ve_first_word (GdmWilling);
+	    if ( ! ve_string_empty (bin) &&
+		 access (bin, X_OK) == 0 &&
+		 (fd = popen (GdmWilling, "r")) != NULL) {
+		    if (fgets (statusBuf, sizeof (statusBuf), fd) != NULL &&
+			! ve_string_empty (g_strstrip (statusBuf))) {
+			    g_free (last_status);
+			    last_status = g_strdup (statusBuf);
+		    } else {
+			    g_free (last_status);
+			    last_status = g_strdup (sysid);
+		    }
+		    pclose (fd);
+	    } else {
+		    g_free (last_status);
+		    last_status = g_strdup (sysid);
+	    }
+	    last_willing = time (NULL);
+	    g_free (bin);
+    }
+
+    if ( ! gdm_is_local_addr (&(clnt_sa->sin_addr)) &&
+	 gdm_xdmcp_displays_from_host (&(clnt_sa->sin_addr)) >= GdmDispPerHost) {
+	    /* Don't translate, this goes over the wire to servers where we
+	     * don't know the charset or language, so it must be ascii */
+	    status.data = g_strdup_printf ("%s (Server is busy)", last_status);
+    } else {
+	    status.data = g_strdup (last_status);
+    }
+    status.length = strlen (status.data);
     
     header.opcode = (CARD16) WILLING;
     header.length = 6 + serv_authlist.authentication.length;
@@ -762,7 +787,10 @@ gdm_xdmcp_send_willing (struct sockaddr_in *clnt_sa)
     XdmcpWriteARRAY8 (&buf, &serv_authlist.authentication); /* Hardcoded authentication */
     XdmcpWriteARRAY8 (&buf, &servhost);
     XdmcpWriteARRAY8 (&buf, &status);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
+    XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		(int)sizeof (struct sockaddr_in));
+
+    g_free (status.data);
 }
 
 static void
@@ -775,7 +803,9 @@ gdm_xdmcp_send_unwilling (struct sockaddr_in *clnt_sa, gint type)
     
     gdm_error (_("Denied XDMCP query from host %s"), inet_ntoa (clnt_sa->sin_addr));
     
-    status.data = _("Display not authorized to connect");
+    /* Don't translate, this goes over the wire to servers where we
+     * don't know the charset or language, so it must be ascii */
+    status.data = "Display not authorized to connect";
     status.length = strlen (status.data);
     
     header.opcode = (CARD16) UNWILLING;
@@ -785,7 +815,98 @@ gdm_xdmcp_send_unwilling (struct sockaddr_in *clnt_sa, gint type)
     
     XdmcpWriteARRAY8 (&buf, &servhost);
     XdmcpWriteARRAY8 (&buf, &status);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
+    XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		(int)sizeof (struct sockaddr_in));
+}
+
+static void
+gdm_xdmcp_really_send_managed_forward (struct sockaddr_in *clnt_sa,
+				       struct sockaddr_in *origin)
+{
+	ARRAY8 address;
+	XdmcpHeader header;
+	struct in_addr addr;
+
+	gdm_debug ("gdm_xdmcp_really_send_managed_forward: "
+		   "Sending MANAGED_FORWARD to %s",
+		   inet_ntoa (clnt_sa->sin_addr));
+
+	address.length = sizeof (struct in_addr);
+	address.data = (void *)&addr;
+	memcpy (address.data, &(origin->sin_addr), sizeof (struct in_addr));
+
+	header.opcode = (CARD16) GDM_XDMCP_MANAGED_FORWARD;
+	header.length = 4 + address.length;
+	header.version = GDM_XDMCP_PROTOCOL_VERSION;
+	XdmcpWriteHeader (&buf, &header);
+
+	XdmcpWriteARRAY8 (&buf, &address);
+	XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		    (int)sizeof (struct sockaddr_in));
+}
+
+static gboolean
+managed_forward_handler (gpointer data)
+{
+	ManagedForward *mf = data;
+	if (gdm_xdmcpfd > 0)
+		gdm_xdmcp_really_send_managed_forward (&(mf->manager),
+						       &(mf->origin));
+	mf->times ++;
+	if (gdm_xdmcpfd <= 0 || mf->times >= 2) {
+		managed_forwards = g_list_remove (managed_forwards, mf);
+		mf->handler = 0;
+		/* mf freed by glib */
+		return FALSE;
+	}
+	return TRUE;
+}
+
+static void
+gdm_xdmcp_send_managed_forward (struct sockaddr_in *clnt_sa,
+				struct sockaddr_in *origin)
+{
+	ManagedForward *mf;
+
+	gdm_xdmcp_really_send_managed_forward (clnt_sa, origin);
+
+	mf = g_new0 (ManagedForward, 1);
+	mf->times = 0;
+	memcpy (&(mf->manager), clnt_sa, sizeof (struct sockaddr_in));
+	memcpy (&(mf->origin), origin, sizeof (struct sockaddr_in));
+
+	mf->handler = g_timeout_add_full (G_PRIORITY_DEFAULT,
+					  MANAGED_FORWARD_INTERVAL,
+					  managed_forward_handler,
+					  mf,
+					  (GDestroyNotify) g_free);
+	managed_forwards = g_list_prepend (managed_forwards, mf);
+}
+
+static void
+gdm_xdmcp_send_got_managed_forward (struct sockaddr_in *clnt_sa,
+				    struct in_addr *origin)
+{
+	ARRAY8 address;
+	XdmcpHeader header;
+	struct in_addr addr;
+
+	gdm_debug ("gdm_xdmcp_send_managed_forward: "
+		   "Sending GOT_MANAGED_FORWARD to %s",
+		   inet_ntoa (clnt_sa->sin_addr));
+
+	address.length = sizeof (struct in_addr);
+	address.data = (void *)&addr;
+	memcpy (address.data, origin, sizeof (struct in_addr));
+
+	header.opcode = (CARD16) GDM_XDMCP_GOT_MANAGED_FORWARD;
+	header.length = 4 + address.length;
+	header.version = GDM_XDMCP_PROTOCOL_VERSION;
+	XdmcpWriteHeader (&buf, &header);
+
+	XdmcpWriteARRAY8 (&buf, &address);
+	XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		    (int)sizeof (struct sockaddr_in));
 }
 
 static char *
@@ -795,7 +916,7 @@ get_host_from_addr (struct sockaddr_in *clnt_sa)
 	struct hostent *he;
 
 	/* Find client hostname */
-	he = gethostbyaddr ((gchar *) &clnt_sa->sin_addr,
+	he = gethostbyaddr ((gchar *) &(clnt_sa->sin_addr),
 			    sizeof (struct in_addr),
 			    AF_INET);
 
@@ -921,9 +1042,12 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
     gdm_debug ("gdm_xdmcp_handle_request: pending=%d, MaxPending=%d, sessions=%d, MaxSessions=%d",
 	       pending, GdmMaxPending, sessions, GdmMaxSessions);
 
+
     /* Check if ok to manage display */
     if (mitauth &&
-	sessions < GdmMaxSessions) {
+	sessions < GdmMaxSessions &&
+	(gdm_is_local_addr (&(clnt_sa->sin_addr)) ||
+	 gdm_xdmcp_displays_from_host (&(clnt_sa->sin_addr)) < GdmDispPerHost)) {
 	    char *disp;
 	    char *hostname = get_host_from_addr (clnt_sa);
 	    disp = g_strdup_printf ("%s:%d", hostname, clnt_dspnum);
@@ -934,14 +1058,23 @@ gdm_xdmcp_handle_request (struct sockaddr_in *clnt_sa, gint len)
 
 	    if (pending >= GdmMaxPending) {
 		    gdm_debug ("gdm_xdmcp_handle_request: maximum pending");
-		    gdm_xdmcp_send_decline (clnt_sa);	
+		    /* Don't translate, this goes over the wire to servers where we
+		    * don't know the charset or language, so it must be ascii */
+		    gdm_xdmcp_send_decline (clnt_sa, "Maximum pending servers");	
 	    } else {
 		    gdm_xdmcp_send_accept (hostname, clnt_sa, clnt_dspnum);
 	    }
 
 	    g_free (hostname);
     } else {
-	    gdm_xdmcp_send_decline (clnt_sa);	
+	    /* Don't translate, this goes over the wire to servers where we
+	    * don't know the charset or language, so it must be ascii */
+	    if ( ! mitauth)
+		    gdm_xdmcp_send_decline (clnt_sa, "Only MIT-MAGIC-COOKIE-1 supported");	
+	    else if (sessions >= GdmMaxSessions)
+		    gdm_xdmcp_send_decline (clnt_sa, "Maximum number of open sessions reached");	
+	    else 
+		    gdm_xdmcp_send_decline (clnt_sa, "Maximum number of open sessions from your host reached");	
     }
 
     XdmcpDisposeARRAY8 (&clnt_authname);
@@ -965,7 +1098,7 @@ gdm_xdmcp_send_accept (const char *hostname,
     ARRAY8 authdata;
     GdmDisplay *d;
     
-    d = gdm_xdmcp_display_alloc (hostname, displaynum);
+    d = gdm_xdmcp_display_alloc (&(clnt_sa->sin_addr), hostname, displaynum);
     
     authentype.data = (CARD8 *) 0;
     authentype.length = (CARD16) 0;
@@ -995,20 +1128,22 @@ gdm_xdmcp_send_accept (const char *hostname,
     XdmcpWriteARRAY8 (&buf, &authname);
     XdmcpWriteARRAY8 (&buf, &authdata);
     
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
+    XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		(int)sizeof (struct sockaddr_in));
     
-    gdm_debug ("gdm_xdmcp_send_accept: Sending ACCEPT to %s with SessionID=%d", 
-	       inet_ntoa (clnt_sa->sin_addr), d->sessionid);
+    gdm_debug ("gdm_xdmcp_send_accept: Sending ACCEPT to %s with SessionID=%ld", 
+	       inet_ntoa (clnt_sa->sin_addr), (long)d->sessionid);
 }
 
 
 static void
-gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa)
+gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa, const char *reason)
 {
     XdmcpHeader header;
     ARRAY8 authentype;
     ARRAY8 authendata;
     ARRAY8 status;
+    GdmForwardQuery *fq;
     
     gdm_debug ("gdm_xdmcp_send_decline: Sending DECLINE to %s", 
 	       inet_ntoa (clnt_sa->sin_addr));
@@ -1019,7 +1154,7 @@ gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa)
     authendata.data = (CARD8 *) 0;
     authendata.length = (CARD16) 0;
     
-    status.data = "Session refused";
+    status.data = (char *)reason;
     status.length = strlen (status.data);
     
     header.version = XDM_PROTOCOL_VERSION;
@@ -1033,7 +1168,16 @@ gdm_xdmcp_send_decline (struct sockaddr_in *clnt_sa)
     XdmcpWriteARRAY8 (&buf, &authentype);
     XdmcpWriteARRAY8 (&buf, &authendata);
     
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
+    XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		(int)sizeof (struct sockaddr_in));
+
+    /* Send MANAGED_FORWARD to indicate that the connection 
+     * reached some sort of resolution */
+    fq = gdm_forward_query_lookup (clnt_sa);
+    if (fq != NULL) {
+	    gdm_xdmcp_send_managed_forward (fq->from_sa, clnt_sa);
+	    gdm_forward_query_dispose (fq);
+    }
 }
 
 
@@ -1045,7 +1189,7 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
     ARRAY8 clnt_dspclass;
     GdmDisplay *d;
     GdmIndirectDisplay *id;
-    gint logfd;
+    GdmForwardQuery *fq;
     
     gdm_debug ("gdm_xdmcp_handle_manage: Got MANAGE from %s", inet_ntoa (clnt_sa->sin_addr));
     
@@ -1068,8 +1212,8 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
 	return;
     }
     
-    gdm_debug ("gdm_xdmcp_handle_manage: Got Display=%d, SessionID=%d from %s", 
-	       clnt_dspnum, clnt_sessid, inet_ntoa (clnt_sa->sin_addr));
+    gdm_debug ("gdm_xdmcp_handle_manage: Got Display=%d, SessionID=%ld from %s", 
+	       (int)clnt_dspnum, (long)clnt_sessid, inet_ntoa (clnt_sa->sin_addr));
     
     /* Display Class */
     if (! XdmcpReadARRAY8 (&buf, &clnt_dspclass)) {
@@ -1078,45 +1222,39 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
     }
     
     d = gdm_xdmcp_display_lookup (clnt_sessid);
-    if (GdmIndirect) {
-	    id = gdm_choose_indirect_lookup (clnt_sa);
-
-	    /* This was an indirect thingie and nothing was yet chosen,
-	     * use a chooser */
-	    if (d != NULL &&
-		d->dispstat == XDMCP_PENDING &&
-		id != NULL &&
-		id->chosen_host == NULL) {
-		    d->use_chooser = TRUE;
-		    d->indirect_id = id->id;
-	    } else {
-		    d->indirect_id = 0;
-		    d->use_chooser = FALSE;
-		    if (id != NULL)
-			    gdm_choose_indirect_dispose (id);
-	    }
-    } else {
-	    d->indirect_id = 0;
-	    d->use_chooser = FALSE;
-    }
-
-    if (d && d->dispstat == XDMCP_PENDING) {
-	gchar *logfile;
+    if (d != NULL &&
+        d->dispstat == XDMCP_PENDING) {
 
 	gdm_debug ("gdm_xdmcp_handle_manage: Looked up %s", d->name);
 
-	/* Log all output from spawned programs to a file */
-	logfile = g_strconcat (GdmLogDir, "/", d->name, ".log", NULL);
-	logfd = open (logfile, O_CREAT|O_TRUNC|O_APPEND|O_WRONLY, 0666);
-	g_free (logfile);
+	if (GdmIndirect) {
+		id = gdm_choose_indirect_lookup (clnt_sa);
 
-	if (logfd != -1) {
-	    dup2 (logfd, 1);
-	    dup2 (logfd, 2);
-            close (logfd);
+		/* This was an indirect thingie and nothing was yet chosen,
+		 * use a chooser */
+		if (d->dispstat == XDMCP_PENDING &&
+		    id != NULL &&
+		    id->chosen_host == NULL) {
+			d->use_chooser = TRUE;
+			d->indirect_id = id->id;
+		} else {
+			d->indirect_id = 0;
+			d->use_chooser = FALSE;
+			if (id != NULL)
+				gdm_choose_indirect_dispose (id);
+		}
+	} else {
+		d->indirect_id = 0;
+		d->use_chooser = FALSE;
+	}
+
+	/* this was from a forwarded query quite apparently so
+	 * send MANAGED_FORWARD */
+	fq = gdm_forward_query_lookup (clnt_sa);
+	if (fq != NULL) {
+		gdm_xdmcp_send_managed_forward (fq->from_sa, clnt_sa);
+		gdm_forward_query_dispose (fq);
 	}
-	else
-	    gdm_error (_("gdm_xdmcp_handle_manage: Could not open logfile for display %s!"), d->name);
 	
 	d->dispstat = XDMCP_MANAGED;
 	sessions++;
@@ -1129,24 +1267,113 @@ gdm_xdmcp_handle_manage (struct sockaddr_in *clnt_sa, gint len)
 	    return;
 	}
     }
-    else if (d && d->dispstat == XDMCP_MANAGED) {
-	gdm_debug ("gdm_xdmcp_handle_manage: Session id %d already managed", clnt_sessid);	
+    else if (d != NULL && d->dispstat == XDMCP_MANAGED) {
+	gdm_debug ("gdm_xdmcp_handle_manage: Session id %ld already managed",
+		   (long)clnt_sessid);	
     }
     else {
-	gdm_debug ("gdm_xdmcp_handle_manage: Failed to look up session id %d", clnt_sessid);
+	gdm_debug ("gdm_xdmcp_handle_manage: Failed to look up session id %ld",
+		   (long)clnt_sessid);
 	gdm_xdmcp_send_refuse (clnt_sa, clnt_sessid);
     }
 
     XdmcpDisposeARRAY8(&clnt_dspclass);
 }
 
+static void 
+gdm_xdmcp_handle_managed_forward (struct sockaddr_in *clnt_sa, gint len)
+{
+	ARRAY8 clnt_address;
+	GdmIndirectDisplay *id;
+
+	gdm_debug ("gdm_xdmcp_handle_managed_forward: "
+		   "Got MANAGED_FORWARD from %s",
+		   inet_ntoa (clnt_sa->sin_addr));
+
+	/* Hostname */
+	if ( ! XdmcpReadARRAY8 (&buf, &clnt_address)) {
+		gdm_error (_("%s: Could not read address"),
+			   "gdm_xdmcp_handle_managed_forward");
+		return;
+	}
+
+	if (clnt_address.length != sizeof (struct in_addr)) {
+		gdm_error (_("%s: Could not read address"),
+			   "gdm_xdmcp_handle_managed_forward");
+		XdmcpDisposeARRAY8 (&clnt_address);
+		return;
+	}
+
+	id = gdm_choose_indirect_lookup_by_chosen
+		(&(clnt_sa->sin_addr), (struct in_addr *)clnt_address.data);
+	if (id != NULL) {
+		gdm_choose_indirect_dispose (id);
+	}
+
+	/* Note: we send GOT even on not found, just in case our previous
+	 * didn't get through and this was a second managed forward */
+	gdm_xdmcp_send_got_managed_forward
+		(clnt_sa, (struct in_addr *)clnt_address.data);
+
+	XdmcpDisposeARRAY8 (&clnt_address);
+}
+
+static void
+gdm_xdmcp_whack_queued_managed_forwards (struct sockaddr_in *clnt_sa,
+					 struct in_addr *origin)
+{
+	GList *li;
+
+	for (li = managed_forwards; li != NULL; li = li->next) {
+		ManagedForward *mf = li->data;
+		if (mf->manager.sin_addr.s_addr == clnt_sa->sin_addr.s_addr &&
+		    mf->origin.sin_addr.s_addr == origin->s_addr) {
+			managed_forwards = g_list_remove_link (managed_forwards, li);
+			g_list_free_1 (li);
+			g_source_remove (mf->handler);
+			/* mf freed by glib */
+			return;
+		}
+	}
+}
+
+static void 
+gdm_xdmcp_handle_got_managed_forward (struct sockaddr_in *clnt_sa, gint len)
+{
+	ARRAY8 clnt_address;
+
+	gdm_debug ("gdm_xdmcp_handle_got_managed_forward: "
+		   "Got MANAGED_FORWARD from %s",
+		   inet_ntoa (clnt_sa->sin_addr));
+
+	/* Hostname */
+	if ( ! XdmcpReadARRAY8 (&buf, &clnt_address)) {
+		gdm_error (_("%s: Could not read address"),
+			   "gdm_xdmcp_handle_got_managed_forward");
+		return;
+	}
+
+	if (clnt_address.length != sizeof (struct in_addr)) {
+		gdm_error (_("%s: Could not read address"),
+			   "gdm_xdmcp_handle_got_managed_forward");
+		XdmcpDisposeARRAY8 (&clnt_address);
+		return;
+	}
+
+	gdm_xdmcp_whack_queued_managed_forwards
+		(clnt_sa, (struct in_addr *)clnt_address.data);
+
+	XdmcpDisposeARRAY8 (&clnt_address);
+}
+
 
 static void
 gdm_xdmcp_send_refuse (struct sockaddr_in *clnt_sa, CARD32 sessid)
 {
     XdmcpHeader header;
+    GdmForwardQuery *fq;
     
-    gdm_debug ("gdm_xdmcp_send_refuse: Sending REFUSE to %d", sessid);
+    gdm_debug ("gdm_xdmcp_send_refuse: Sending REFUSE to %ld", (long)sessid);
     
     header.version = XDM_PROTOCOL_VERSION;
     header.opcode= (CARD16) REFUSE;
@@ -1154,7 +1381,16 @@ gdm_xdmcp_send_refuse (struct sockaddr_in *clnt_sa, CARD32 sessid)
     
     XdmcpWriteHeader (&buf, &header);  
     XdmcpWriteCARD32 (&buf, sessid);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
+    XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		(int)sizeof (struct sockaddr_in));
+
+    /* this was from a forwarded query quite apparently so
+    * send MANAGED_FORWARD */
+    fq = gdm_forward_query_lookup (clnt_sa);
+    if (fq != NULL) {
+	    gdm_xdmcp_send_managed_forward (fq->from_sa, clnt_sa);
+	    gdm_forward_query_dispose (fq);
+    }
 }
 
 
@@ -1164,8 +1400,10 @@ gdm_xdmcp_send_failed (struct sockaddr_in *clnt_sa, CARD32 sessid)
     XdmcpHeader header;
     ARRAY8 status;
     
-    gdm_debug ("gdm_xdmcp_send_failed: Sending FAILED to %d", sessid);
+    gdm_debug ("gdm_xdmcp_send_failed: Sending FAILED to %ld", (long)sessid);
     
+    /* Don't translate, this goes over the wire to servers where we
+     * don't know the charset or language, so it must be ascii */
     status.data = "Failed to start session";
     status.length = strlen (status.data);
     
@@ -1176,7 +1414,8 @@ gdm_xdmcp_send_failed (struct sockaddr_in *clnt_sa, CARD32 sessid)
     XdmcpWriteHeader (&buf, &header);
     XdmcpWriteCARD32 (&buf, sessid);
     XdmcpWriteARRAY8 (&buf, &status);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
+    XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		(int)sizeof (struct sockaddr_in));
 }
 
 
@@ -1217,7 +1456,7 @@ gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD32 sessid)
 {
     XdmcpHeader header;
     
-    gdm_debug ("Sending ALIVE to %d", sessid);
+    gdm_debug ("Sending ALIVE to %ld", (long)sessid);
     
     header.version = XDM_PROTOCOL_VERSION;
     header.opcode = (CARD16) ALIVE;
@@ -1226,7 +1465,8 @@ gdm_xdmcp_send_alive (struct sockaddr_in *clnt_sa, CARD32 sessid)
     XdmcpWriteHeader (&buf, &header);
     XdmcpWriteCARD8 (&buf, 1);
     XdmcpWriteCARD32 (&buf, sessid);
-    XdmcpFlush (xdmcpfd, &buf, clnt_sa, (int)sizeof (struct sockaddr_in));
+    XdmcpFlush (gdm_xdmcpfd, &buf, (XdmcpNetaddr)clnt_sa,
+		(int)sizeof (struct sockaddr_in));
 }
 
 
@@ -1259,7 +1499,7 @@ gdm_xdmcp_host_allow (struct sockaddr_in *clnt_sa)
 
 
 static GdmDisplay *
-gdm_xdmcp_display_alloc (const char *hostname, gint displaynum)
+gdm_xdmcp_display_alloc (struct in_addr *addr, const char *hostname, gint displaynum)
 {
     GdmDisplay *d = NULL;
     
@@ -1275,14 +1515,22 @@ gdm_xdmcp_display_alloc (const char *hostname, gint displaynum)
     d->sesspid = 0;
     d->slavepid = 0;
     d->type = TYPE_XDMCP;
+    d->console = FALSE;
     d->dispstat = XDMCP_PENDING;
     d->sessionid = globsessid++;
     d->acctime = time (NULL);
     d->dispnum = displaynum;
 
+#ifdef __linux__
+    d->vt = -1;
+#endif
+
+    d->logged_in = FALSE;
+    d->login = NULL;
+
     d->sleep_before_run = 0;
     if (GdmAllowRemoteAutoLogin &&
-	! gdm_string_empty (GdmTimedLogin)) {
+	! ve_string_empty (GdmTimedLogin)) {
 	    d->timed_login_ok = TRUE;
     } else {
 	    d->timed_login_ok = FALSE;
@@ -1291,6 +1539,7 @@ gdm_xdmcp_display_alloc (const char *hostname, gint displaynum)
     d->name = g_strdup_printf ("%s:%d", hostname,
 			       displaynum);
     d->hostname = g_strdup (hostname);
+    memcpy (&d->addr, addr, sizeof (struct in_addr));
     
     /* Secure display with cookie */
     if (! gdm_auth_secure_display (d))
@@ -1300,8 +1549,8 @@ gdm_xdmcp_display_alloc (const char *hostname, gint displaynum)
     
     pending++;
     
-    gdm_debug ("gdm_xdmcp_display_alloc: display=%s, session id=%d, pending=%d ",
-	       d->name, d->sessionid, pending);
+    gdm_debug ("gdm_xdmcp_display_alloc: display=%s, session id=%ld, pending=%d ",
+	       d->name, (long)d->sessionid, pending);
     
     return (d);
 }
@@ -1374,8 +1623,8 @@ gdm_xdmcp_displays_check (void)
 		    d->type == TYPE_XDMCP &&
 		    d->dispstat == XDMCP_PENDING &&
 		    time (NULL) > d->acctime + GdmMaxManageWait) {
-			gdm_debug ("gdm_xdmcp_displays_check: Disposing session id %d",
-				   d->sessionid);
+			gdm_debug ("gdm_xdmcp_displays_check: Disposing session id %ld",
+				   (long)d->sessionid);
 			gdm_display_dispose (d);
 
 			/* restart as the list is now fucked */
@@ -1411,5 +1660,4 @@ gdm_xdmcp_close (void)
 
 #endif /* HAVE_LIBXDMCP */
 
-
 /* EOF */
