@@ -18,14 +18,38 @@
  * 
  */
 
-#include <config.h>
+#include "config.h"
 #include <gnome.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <errno.h>
+
+#include <viciousui.h>
+
 #include "gdm.h"
 
+/* If path starts with a "trusted" directory, don't sanity check things */
+static gboolean
+is_in_trusted_pic_dir (const char *path)
+{
+	char *globalpix;
+
+	/* our own pixmap dir is trusted */
+	if (strncmp (path, EXPANDED_PIXMAPDIR, sizeof (EXPANDED_PIXMAPDIR)) == 0)
+		return TRUE;
+
+	/* gnome's pixmap dir is trusted */
+	globalpix = gnome_unconditional_pixmap_file ("");
+	if (strncmp (path, globalpix, strlen (globalpix)) == 0) {
+		g_free (globalpix);
+		return TRUE;
+	}
+	g_free (globalpix);
+
+	return FALSE;
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -55,7 +79,6 @@ main (int argc, char *argv[])
 					    "it\nin the GDM configurator "
 					    "program."));
 		gnome_dialog_run_and_close (GNOME_DIALOG (d));
-		return 0;
 	}
 
 	dialog = gnome_dialog_new (_("Select a photo"),
@@ -73,7 +96,7 @@ main (int argc, char *argv[])
 	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (dialog)->vbox),
 			    photo, TRUE, TRUE, 0);
 
-	if ( ! gdm_string_empty (last_pix)) {
+	if ( ! ve_string_empty (last_pix)) {
 		GtkWidget *e = gnome_pixmap_entry_gtk_entry
 			(GNOME_PIXMAP_ENTRY (photo));
 		gtk_entry_set_text (GTK_ENTRY (e), last_pix);
@@ -84,11 +107,25 @@ main (int argc, char *argv[])
 	while (gnome_dialog_run (GNOME_DIALOG (dialog)) == 0) {
 		struct stat s;
 		char *pixmap = gnome_pixmap_entry_get_filename (GNOME_PIXMAP_ENTRY (photo));
-		if (gdm_string_empty (pixmap) ||
+		if (ve_string_empty (pixmap) ||
 		    stat (pixmap, &s) < 0) {
 			GtkWidget *d;
 			d = gnome_warning_dialog (_("No picture selected."));
 			gnome_dialog_run_and_close (GNOME_DIALOG (d));
+		} else if (is_in_trusted_pic_dir (pixmap)) {
+			/* Picture is in trusted dir, no need to copy nor
+			 * check it */
+
+			/* yay, leak, who cares */
+			char *cfg_file = g_strconcat (g_get_home_dir (), 
+						      "/.gnome/gdm",
+						      NULL);
+			gnome_config_set_string ("/gdm/face/picture",
+						 pixmap);
+			gnome_config_sync ();
+			/* ensure proper permissions */
+			chmod (cfg_file, 0600);
+			break;
 		} else if (s.st_size > max_size) {
 			char *msg;
 			GtkWidget *d;
@@ -107,6 +144,9 @@ main (int argc, char *argv[])
 			char *photofile = g_strconcat (g_get_home_dir (), 
 						       "/.gnome/photo",
 						       NULL);
+			char *cfg_file = g_strconcat (g_get_home_dir (), 
+						      "/.gnome/gdm",
+						      NULL);
 			int fddest, fdsrc;
 
 			fdsrc = open (pixmap, O_RDONLY);
@@ -119,9 +159,11 @@ main (int argc, char *argv[])
 					 g_strerror (errno));
 				d = gnome_warning_dialog (msg);
 				gnome_dialog_run_and_close (GNOME_DIALOG (d));
+				g_free (cfg_file);
 				g_free (photofile);
 				continue;
 			}
+			unlink (photofile);
 			fddest = open (photofile, O_WRONLY | O_CREAT);
 			if (fddest < 0) {
 				GtkWidget *d;
@@ -132,6 +174,7 @@ main (int argc, char *argv[])
 					 g_strerror (errno));
 				d = gnome_warning_dialog (msg);
 				gnome_dialog_run_and_close (GNOME_DIALOG (d));
+				g_free (cfg_file);
 				g_free (photofile);
 				close (fdsrc);
 				continue;
@@ -144,7 +187,10 @@ main (int argc, char *argv[])
 			close (fddest);
 			gnome_config_set_string ("/gdmphotosetup/last/picture",
 						 pixmap);
+			gnome_config_set_string ("/gdm/face/picture", "");
 			gnome_config_sync ();
+			/* ensure proper permissions */
+			chmod (cfg_file, 0600);
 			break;
 		}
 	}
